from mirmod.utils.logger import logger
from ..security.security_context import Security_context
from ..orm.base_orm import Base_object_ORM
import importlib.util
import sys
import json
from mirmod.execution_context import get_execution_context
from mirmod.platform_versions import PLATFORM_VERSION
import os
import copy


def _load_plugin(plugin_path, plugin_name):
    spec = importlib.util.spec_from_file_location(plugin_name, plugin_path)
    plugin_module = importlib.util.module_from_spec(spec)
    sys.modules[plugin_name] = plugin_module
    spec.loader.exec_module(plugin_module)
    return plugin_module


class model_card_api:
    def load_artefacts(self, destination_path: str = ""):
        """Load artefacts from the model card database and download to the local file system."""
        assert False, "load_artefacts() not implemented"

    def predict(self, features):
        """Make a prediction and return a result set"""
        assert False, "predict() not implemented"

    def train(
        self,
        features=None,
        labels=None,
        test_features=None,
        test_labels=None,
        callback_function=None,
    ):
        """Train the model and on every epoc call the callback function with the epoc number, loss and accuracy."""
        assert False, "train() not implemented"

    def save_artefacts(self, source_path: str = "", loss=0.0, nepochs=1):
        """Save artefacts to the model card database."""
        assert False, "save_artefacts() not implemented"

    def delete_artefacts(self):
        """Delete artefacts from the model card database. Note that this doesn't remove the model card entry."""
        assert False, "delete_artefacts() not implemented"

    def list_artefacts(self):
        """List all artefacts that are stored in the model card database."""
        assert False, "list_artefacts() not implemented"

    def create_new_model_impl(self, params: dict):
        """Every model frame work has their own unique way to create model instances. The parameters for creating the model are passed in as a dictionary."""
        assert False, "create_new_model_impl() not implemented"

    def get_model_implementation(self):
        """Returns the underlying model implementation which allows for direct access to it. The type of this object is determined by the underlying framework used."""
        assert False, "get_model_implementation() not implemented"

    def set_model_card(self, model):
        """Sets the model card object."""
        assert False, "set_model_card() not implemented"

    def get_optimizer(self):
        """Returns the optimizer used for training the model."""
        assert False, "get_optimizer() not implemented"

    def set_optimizer(self, i):
        """Sets the optimizer used for training the model."""
        assert False, "set_optimizer() not implemented"

    def get_criterion(self):
        """Returns the criterion used for estimating loss."""
        assert False, "get_criterion() not implemented"

    def set_criterion(self, i):
        """Sets the criterion used for estimating loss."""
        assert False, "set_criterion() not implemented"


class Model(Base_object_ORM, model_card_api):
    # Model table has alias t
    sql_object_ORM = {
        "id": "t.id as id",
        "model_type": "t.model_type as model_type",
        "miranda_version": "t.miranda_version as miranda_version",
        "hardware": "t.hardware as hardware",
        "load_url": "t.load_url as load_url",
        "save_url": "t.save_url as save_url",
        "files": "t.files as files",
        "authors": "t.authors as authors",
        "license": "t.license as license",
        "feature_labels": "t.feature_labels as feature_labels",
        "prediction_labels": "t.prediction_labels as prediction_labels",
        "feature_units": "t.feature_units as feature_units",
        "prediction_units": "t.prediction_units as prediction_units",
        "loss": "t.loss as loss",
        "accuracy": "t.accuracy as accuracy",
        "precision": "t.precision as `precision`",
        "recall": "t.recall as recall",
        "knowledge_object_id": "t.knowledge_object_id as knowledge_object_id",
        "workflow_state": "t.workflow_state as workflow_state",
    }

    sql_object_ORM.update(Base_object_ORM.metadata)

    def __init__(
        self, sc: Security_context, id=-1, metadata_id=None, user_id=-1, model_type=None
    ):
        self.default_value = {
            "id": -1,
            "metadata_id": -1,
            "model_type": "PYTORCH",
            "miranda_version": PLATFORM_VERSION,
            "hardware": '["CPU"]',
            "load_url": "",
            "save_url": "",
            "files": "[]",
            "authors": "[]",
            "license": "",
            "feature_labels": "[]",
            "prediction_labels": "[]",
            "feature_units": "[]",
            "prediction_units": "[]",
            "loss": 0.0,
            "accuracy": 0.0,
            "precision": 0.0,
            "recall": 0.0,
            "knowledge_object_id": -1,
            "workflow_state": "UNINITIALIZED",
        }
        self.has_loaded_artefacts = False
        self.id = id
        self.sctx = sc
        self.create_mapping(self.sql_object_ORM, "model")
        self.model_type = model_type
        self.model_impl: model_card_api = None
        if metadata_id is not None:
            self._load_from_metadata_id(sc, metadata_id, user_id=user_id)
            # logger.info("Loaded ML_model object with metadata id: " + str(metadata_id))
        elif id != -1:
            self._load_from_id(sc, self.id)
            # logger.info("Loaded ML_model object with id: " + str(id))
        self.params = None

        try:
            ecx = get_execution_context()
            self.storage_policy = ecx.get_storage_policy()
        except Exception as e:
            logger.error(f"Error getting storage policy: {e}")
            self.storage_policy = None

    def load_artefacts(self, destination_path: str = ""):
        """Load artefacts from the model card database and download to the local file system."""
        if self.has_loaded_artefacts:
            print("|=> load_artefact: Artefacts already loaded.")
            return
        if self.model_impl is None:
            print("|=> load_artefact: No model implementation has been specified.")
            return
        self.has_loaded_artefacts = True
        if self.storage_policy is not None:
            files = json.loads(self.files)
            for f in files:
                self.storage_policy.load_file(os.path.join(destination_path, f))
        self.model_impl.load_artefacts(destination_path)

    def predict(self, features):
        """Make a prediction and return a result set"""
        assert self.has_loaded_artefacts, (
            "Artefacts must be loaded before making a prediction."
        )
        return self.model_impl.predict(features)

    def train(
        self,
        features=None,
        labels=None,
        test_features=None,
        test_labels=None,
        callback_function=None,
    ):
        """Train the model and on every epoch call the callback function with the epoch number, loss and accuracy.
        For every epoch the call_backfunction is called with the following parameters:
        callback_function(epoch, loss, accuracy)
        """
        ret = self.model_impl.train(
            features, labels, test_features, test_labels, callback_function
        )
        self.has_loaded_artefacts = True
        return ret

    def save_artefacts(self, source_path: str = "", loss=0.0, nepochs=1):
        """Save artefacts to the model card database."""
        # First we call the model implementation to save any artefacts to the local file system
        self.model_impl.save_artefacts(source_path, loss=loss, nepochs=nepochs)
        # We ask the model implementation to supply us with the filenames of the artefacts
        files = self.model_impl.list_artefacts()
        self.files = json.dumps(files)
        if files is None:
            return
        for f in files:
            # We use the storage policy to save the artefacts to a persistent storage
            self.storage_policy.save_file(os.path.join(source_path, f))
            self.load_url = self.storage_policy.load_url
            self.save_url = self.storage_policy.save_url
        # We update the database with the new list of files
        self.update(self.sctx)

    def copy_to(self, new_name: str, new_description: str, path=""):
        self.load_artefacts()
        # get the parameters needed to recreate the model implementation
        try:
            id, _ = next(find_model_id_by_name(self.sctx, new_name))
            new_model: Model = Model(self.sctx, id=id)
            new_model.has_loaded_artefacts = True
        except StopIteration:
            new_model: Model = Model(self.sctx, id=-1)
            with self.sctx.connect() as con:
                id, _ = new_model.create(con, new_name, new_description)
                new_model = Model(self.sctx, id=id)
            assert new_model.id != -1, "Target '{}' could not be created.".format(
                new_name
            )
        new_model.model_impl = copy.copy(self.model_impl)
        new_model.model_impl.set_model_card(new_model)
        self.model_impl.save_artefacts(
            path
        )  # make sure the artefacts are saved to local disk
        for f1, f2 in zip(
            self.model_impl.list_artefacts(), new_model.model_impl.list_artefacts()
        ):
            print("cp {} {}".format(os.path.join(path, f1), os.path.join(path, f2)))
            os.system("cp {} {}".format(os.path.join(path, f1), os.path.join(path, f2)))
        new_model.has_loaded_artefacts = True
        if self.storage_policy is not None:
            for f in new_model.model_impl.list_artefacts():
                self.storage_policy.save_file(os.path.join(path, f))
            files = new_model.model_impl.list_artefacts()
            new_model.files = json.dumps(files)
        new_model.set_feature_labels(self.get_feature_labels())
        new_model.set_prediction_labels(self.get_prediction_labels())
        new_model.set_feature_units(self.get_feature_units())
        new_model.set_prediction_units(self.get_prediction_units())
        new_model.update(self.sctx)
        return new_model

    def get_model_implementation(self):
        # This actually returns the implementation of the model of the instance of the framework, so effectively the implementation of the implementation.
        return self.model_impl.get_model_implementation()

    def delete_artefacts(self):
        """Delete artefacts from the model card database. This call does nothing for the local files."""
        self.delete_artefacts()

    def get_feature_labels(self):
        if self.feature_labels is None:
            return []
        if isinstance(self.feature_labels, list):
            return self.feature_labels
        return json.loads(self.feature_labels)

    def set_feature_labels(self, labels):
        if isinstance(labels, list):
            self.feature_labels = json.dumps(labels)
        elif isinstance(labels, str):
            self.feature_labels = labels
        else:
            raise Exception("Labels must be a list or a string")

    def set_prediction_labels(self, labels):
        if isinstance(labels, list):
            self.prediction_labels = json.dumps(labels)
        elif isinstance(labels, str):
            self.prediction_labels = labels
        else:
            raise Exception("Labels must be a list or a string")

    def get_prediction_labels(self):
        if self.prediction_labels is None:
            return []
        if isinstance(self.prediction_labels, list):
            return self.prediction_labels
        return json.loads(self.prediction_labels)

    def get_feature_units(self):
        if self.feature_units is None:
            return []
        if isinstance(self.feature_units, list):
            return self.feature_units
        return json.loads(self.feature_units)

    def set_feature_units(self, units):
        if isinstance(units, list):
            self.feature_units = json.dumps(units)
        elif isinstance(units, str):
            self.feature_units = units
        else:
            raise Exception("Units must be a list or a string")

    def set_prediction_units(self, units):
        if isinstance(units, list):
            self.prediction_units = json.dumps(units)
        elif isinstance(units, str):
            self.prediction_units = units
        else:
            raise Exception("Units must be a list or a string")

    def get_prediction_units(self):
        if self.prediction_units is None:
            return []
        if isinstance(self.prediction_units, list):
            return self.prediction_units
        return json.loads(self.prediction_units)

    def get_optimizer(self):
        """Returns the optimizer used for training the model."""
        return self.model_impl.get_optimizer()

    def set_optimizer(self, i):
        """Sets the optimizer used for training the model."""
        self.model_impl.set_optimizer(i)

    def get_criterion(self):
        """Returns the criterion used for estimating loss."""
        return self.model_impl.get_criterion()

    def set_criterion(self, i):
        """Sets the criterion used for estimating loss."""
        self.model_impl.set_criterion(i)


# ---------------- END class Model ----------------


def find_model_id_by_name(sc: Security_context, name: str):
    with sc.connect() as conn:
        with conn.cursor() as cur:
            sql = "SELECT t.id, t.model_type FROM v_model t WHERE t.name = %s"
            cur.execute(sql, (name,))
            rows = cur.fetchall()
            for row in rows:
                yield row


def create_or_update_modelcard(
    sctx,
    name,
    description,
    model_type="PYTORCH",
    miranda_version=PLATFORM_VERSION,
    hardware=["GPU"],
    authors="MainlyAI",
    license="",
    prediction_units=[],
    feature_units=[],
    prediction_labels=[],
):
    """Attempt to create a new model card. If a model card with the same name already exists it is loaded, modified and returned."""
    found = False
    model_card = Model(sctx, id=-1, model_type="PYTORCH")
    for m in find_model_id_by_name(sctx, name):
        if m[1] != model_type:
            raise Exception(
                "Model card with name "
                + name
                + " already exists with a different model type"
            )
        print(
            "Found an existing model using name '{}' (id={}). Load all the data from the database.".format(
                name, m[0]
            )
        )
        model_card = Model(sctx, id=m[0])
        assert id != -1, (
            "Model card with name "
            + name
            + " was found with id = {} but could not be loaded.".format(m[0])
        )
        found = True
        break

    if not found:
        with sctx.connect() as con:
            model_id, metadata_id = model_card.create(
                con, name, description, model_type=model_type
            )
            con.commit()
            model_card = Model(sctx, id=model_id)
            print(
                "Created a new model card entry with ID: {} ({})".format(
                    model_id, metadata_id
                )
            )
    model_card.authors = authors
    model_card.model_type = model_type
    model_card.miranda_version = miranda_version
    model_card.hardware = json.dumps(hardware)
    model_card.license = license
    model_card.prediction_units = json.dumps(prediction_units)
    model_card.feature_units = json.dumps(feature_units)
    model_card.prediction_labels = json.dumps(prediction_labels)
    model_card.update(sctx)
    return model_card



from contextlib import redirect_stderr, redirect_stdout
from mirmod.miranda_exceptions import MirandaStopCurrentIterator
from io import StringIO
import json
import os
import traceback
import re

from mirmod import miranda


def analyze_stacktrace(st, wob_name):
    def error_on_object_attribute(stack_trace):
        stack_trace_no_newlines = stack_trace.replace("\n", "")
        pattern = re.compile(
            r'File ".+?/workflow_object\.py", line \d+.*?'
            r"AttributeError: \'(?P<object>\w+)\' object has no attribute \'(?P<attribute>\w+)\'"
        )
        match = pattern.search(stack_trace_no_newlines)
        if match:
            return match.group("object"), match.group("attribute")
        return None, None

    object, attribute = error_on_object_attribute(st)
    if object and attribute:
        return "It seems you didn't initialize the variable '{}' in the method 'init' of the workflow object '{}'.\nTry something like: \n\tself.{} = None\n".format(
            attribute, wob_name, attribute
        )

    return st


ValidTypes = [
    "dataframe",
    "modelcard",
    "hfdataset",
    "string",
    "number",
    "image",
    "plotly",
    "document",
    "unit_dataframe",
    "data",
    "value",
    "model",
    "state",
]

DeprecatedTypes = [
    "dataframe",
    "modelcard",
    "hfdataset",
    "string",
    "number",
    "image",
    "plotly",
    "document",
    "unit_dataframe",
]


class Attribute:
    def __init__(
        self,
        parent,
        kind: str,
        name=None,
        control=None,
        hidden=False,
        connectable=True,
        edge_text="",
        recommendations=[],
        shown_when={},
        group="",
    ):
        if kind.lower() not in ValidTypes:
            raise Exception(f'Invalid attribute kind "{kind}"')

        # if kind.lower() in DeprecatedTypes:
        #   print(f'WARNING: Attribute kind "{kind}" is deprecated. Please use the new kinds: "value", "data", "model", "state"')

        self.parent = parent
        self.kind = kind.lower()
        self.name = name
        self.control = control
        self.hidden = hidden
        self.connectable = connectable
        self.edge_text = edge_text
        self.recommendations = recommendations
        self.direction = None
        self.shown_when = shown_when
        self.group = group


class Transmitter(Attribute):
    def __init__(
        self,
        parent,
        kind,
        name,
        transmit,
        control=None,
        hidden=False,
        connectable=True,
        edge_text="",
        recommendations=[],
        shown_when={},
        group="",
    ):
        super().__init__(
            parent,
            kind,
            name,
            control,
            hidden,
            connectable,
            edge_text,
            recommendations,
            shown_when=shown_when,
            group=group,
        )
        self.transmit = transmit
        self.direction = miranda.TRANSMITTER


class Transmitter_field(Transmitter):
    def __init__(
        self,
        parent,
        kind,
        name,
        transmit,
        control=None,
        hidden=False,
        connectable=True,
        edge_text="",
        recommendations=[],
        shown_when={},
        group="",
    ):
        super().__init__(
            parent,
            kind,
            name,
            transmit,
            control,
            hidden,
            connectable,
            edge_text,
            recommendations,
            shown_when=shown_when,
            group=group,
        )
        self.direction = miranda.TRANSMITTER_FIELD


class Receiver(Attribute):
    def __init__(
        self,
        parent,
        kind,
        name,
        receive,
        control=None,
        hidden=False,
        connectable=True,
        edge_text="",
        recommendations=[],
        shown_when={},
        group="",
    ):
        super().__init__(
            parent,
            kind,
            name,
            control=control,
            hidden=hidden,
            connectable=connectable,
            edge_text=edge_text,
            recommendations=recommendations,
            shown_when=shown_when,
            group=group,
        )
        self.receive = receive
        self.direction = miranda.RECEIVER


class Receiver_field(Receiver):
    def __init__(
        self,
        parent,
        kind,
        name,
        receive,
        control=None,
        hidden=False,
        connectable=True,
        edge_text="",
        recommendations=[],
        shown_when={},
        group="",
    ):
        super().__init__(
            parent,
            kind,
            name,
            receive,
            control,
            hidden,
            connectable,
            edge_text,
            recommendations,
            shown_when=shown_when,
            group=group,
        )
        self.direction = miranda.RECEIVER_FIELD


class Passthrough(Attribute):
    def __init__(
        self,
        parent,
        kind,
        name,
        ts_func,
        rc_func,
        recommendations={},
        shown_when={},
        group="",
    ):
        super().__init__(
            parent,
            kind,
            name,
            recommendations=recommendations,
            shown_when=shown_when,
            group=group,
        )
        self.transmit = ts_func
        self.receive = rc_func
        self.direction = miranda.PASSTHROUGH


def default_init(self):
    pass


def default_execute(self):
    pass


def default_setup(self):
    pass


class WOB:
    def __init__(self, name: str = None):
        self.name = name
        self.attributes: dict[str, Attribute] = {}
        self.executed = {"errors": [], "warnings": [], "has_executed": False}
        self._init = default_init
        self._execute = default_execute
        self._setup = default_setup
        self.status = self.Status(self.update_status)

    def get(self, name):
        if name not in self.attributes:
            raise Exception("Attribute does not exist")
        return self.attributes[name]

    def set(self, name, value):
        if name not in self.attributes:
            raise Exception("Attribute does not exist")
        self.attributes[name].set(value)

    def init(self, name=None):
        if name is not None:
            self.name = name
        if self._init is not None and self._init != default_init:
            raise Exception("WOB already initialized")

        def decorator(func):
            self._init = func
            return self._init

        return decorator

    def transmitter(
        self,
        kind,
        name,
        control=None,
        is_field=False,
        hidden=False,
        connectable=True,
        edge_text="",
        recommendations=[],
        shown_when={},
        group="",
    ):
        def decorator(func):
            def transmit(self):
                # assert target == self.key
                try:
                    rs = func(self)
                    return rs
                except Exception as e:
                    # send stacktrace to the analyzer
                    stacktrace = traceback.format_exc()
                    print("ERROR: ", e)
                    raise Exception(analyze_stacktrace(stacktrace, self.name))

            if is_field:
                newAttribute = Transmitter_field(
                    self,
                    kind,
                    name,
                    transmit,
                    control,
                    hidden,
                    connectable,
                    edge_text,
                    recommendations,
                    shown_when=shown_when,
                    group=group,
                )
            else:
                newAttribute = Transmitter(
                    self,
                    kind,
                    name,
                    transmit,
                    control,
                    hidden,
                    connectable,
                    edge_text,
                    recommendations,
                    shown_when=shown_when,
                    group=group,
                )
            if name in self.attributes:
                raise Exception("Attribute already exists")
            self.attributes[name] = newAttribute
            return func

        return decorator

    def receiver(
        self,
        kind,
        name,
        default=None,
        control=None,
        is_field=False,
        hidden=False,
        connectable=True,
        edge_text="",
        recommendations=[],
        shown_when={},
        group="",
    ):
        def decorator(func):
            def receive(self, value):
                # assert value == self.key
                try:
                    rs = func(self, value)
                    return rs
                except MirandaStopCurrentIterator:
                    raise MirandaStopCurrentIterator  # pass along to next layer
                except Exception as e:
                    # send stacktrace to the analyzer
                    stacktrace = traceback.format_exc()
                    print("ERROR: ", e)
                    raise Exception(analyze_stacktrace(stacktrace, self.name))

            if is_field:
                newAttribute = Receiver_field(
                    self,
                    kind,
                    name,
                    receive,
                    control,
                    hidden,
                    connectable,
                    edge_text,
                    recommendations,
                    shown_when=shown_when,
                    group=group,
                )
            else:
                newAttribute = Receiver(
                    self,
                    kind,
                    name,
                    receive,
                    control,
                    hidden,
                    connectable,
                    edge_text,
                    recommendations,
                    shown_when=shown_when,
                    group=group,
                )
            if name in self.attributes:
                raise Exception("Attribute already exists")
            self.attributes[name] = newAttribute
            return func

        return decorator

    def passthrough(
        self, kind, name, internal_name, recommendations={}, shown_when={}, group=""
    ):
        def ts_func(self):
            return getattr(self, internal_name, None)

        def rc_func(self, value):
            setattr(self, internal_name, value)

        newAttribute = Passthrough(
            self,
            kind,
            name,
            ts_func,
            rc_func,
            recommendations,
            shown_when=shown_when,
            group=group,
        )
        self.attributes[name] = newAttribute

    def execute(self):
        if self._execute is not None and self._execute != default_execute:
            raise Exception("WOB already has execute")

        def decorator(func):
            self._execute = func
            return self._execute

        return decorator

    def setup(self):
        if self._setup is not None and self._setup != default_setup:
            raise Exception("WOB already has setup")

        def decorator(func):
            self._setup = func
            return self._setup

        return decorator

    def update_status(self, payload, is_global=False):
        ecx = miranda.get_execution_context()
        sc = ecx.get_security_context()
        ob = None
        if is_global:
            ob = ecx.get_knowledge_object()
        else:
            ob = ecx.get_current_wob()
        print(f"|=> Updating status: {ob.id} {payload}")
        if ob is None or ob.id == -1:
            print("|=> WARNING: Failed to update status. No such object exists.")
            return
        ob.status = json.dumps(payload)
        ob.update(sc)
        miranda.notify_gui(
            sc,
            json.dumps(
                {
                    "action": "status",
                    "data": {
                        "id": ob.id,
                        "metadata_id": ob.metadata_id,
                        "status": payload,
                    },
                }
            ),
        )
        if payload["timeout"] is not None:
            # quietly remove the status during the timeout to ensure it doesnt linger, as timeout is client-only
            ob.status = "null"
            ob.update(sc)

    class Status:
        def __init__(self, update_fn):
            self.update_fn = update_fn

        def progress(
            self,
            name,
            value,
            max_value=0,
            icon="",
            color="",
            timeout=None,
            is_global=False,
        ):
            print(f"|=> PROGRESS: {name}: {value}/{max_value}")
            self.update_fn(
                {
                    "kind": "progress",
                    "name": name,
                    "value": value,
                    "max_value": max_value,
                    "icon": icon,
                    "color": color,
                    "timeout": timeout,
                },
                is_global,
            )

        def info(self, name, icon="", color="", timeout=None, is_global=False):
            self.update_fn(
                {
                    "kind": "info",
                    "name": name,
                    "icon": icon,
                    "color": color,
                    "timeout": timeout,
                },
                is_global,
            )

        def waiting(self, name, icon="", color="", timeout=None, is_global=False):
            self.update_fn(
                {
                    "kind": "waiting",
                    "name": name,
                    "icon": icon,
                    "color": color,
                    "timeout": timeout,
                },
                is_global,
            )

        def clear(self, is_global=False):
            self.update_fn("null", is_global)


def _unsafe_get_code_entry_class(
    program_source: str, entry_class_type=WOB, program_name: str = "wob.py"
):
    """
    Returns the entry class of the program source by executing it in an unprotected context then matching all locals against entry_class_type.
    WARNING: This function evaluates program_source in an unprotected context. This should only be used in temporary execution contexts.
    """

    if "I_AM_IN_AN_ISOLATED_AND_SAFE_CONTEXT" not in os.environ:
        raise Exception(
            "This function should only be used in temporary execution contexts, as it executes arbitrary user provided code. To acknowledge this, set the environment variable I_AM_IN_AN_ISOLATED_AND_SAFE_CONTEXT to 1."
        )

    stdout = StringIO()
    stderr = StringIO()

    entry_class = None

    with redirect_stdout(stdout), redirect_stderr(stderr):
        program = compile(program_source, program_name, "exec")
        program_locals = {}
        program_globals = {}
        exec(program, program_globals, program_locals)
        for k, v in program_locals.items():
            try:
                if isinstance(v, entry_class_type):
                    entry_class = v
                    break
                pass
            except Exception:
                pass  # we're just interested in entry_class and we can safely ignore weird errors

    return (entry_class, stdout, stderr, program_globals, program_locals)



from ..orm.base_orm import Base_object_ORM
from ..security.security_context import Security_context


class Project(Base_object_ORM):
    """A project may link several knowledge objects (graphs) together."""

    sql_project_ORM = {
        "id": "t.id as id",
        "organization_id": "t.organization_id as organization_id",
        "stripe_sub_id": "t.stripe_sub_id as stripe_sub_id",
        "last_paid": "t.last_paid as last_paid",
    }
    sql_project_ORM.update(Base_object_ORM.metadata)

    def __init__(self, sc: Security_context, id=-1, metadata_id=None, user_id=-1):
        self.default_value = {
            "organization_id": -1,
            "stripe_sub_id": "",
            "last_paid": "1970-01-01 00:00:00",
        }
        self.buffer = {}  # Storing execution context data. TODO will hook up with library for thread context data
        self.id = id
        self.sctx = sc
        self.create_mapping(self.sql_project_ORM, "project")
        if metadata_id is not None:
            self._load_from_metadata_id(sc, metadata_id, user_id=user_id)
        elif id != -1:
            self._load_from_id(sc, self.id)



from ..orm.base_orm import Base_object_ORM
from ..security.security_context import Security_context


class Deployment(Base_object_ORM):
    """The deployment object, like the Docker_job object is built for the purpose of storing the
    information about the deployment of a workflow object."""

    sql_ORM = {
        "id": "t.id as id",
        "public": "t.public as public",
        "workflow_state": "t.workflow_state as workflow_state",
        "deployed_at": "t.deployed_at as deployed_at",
        "deployed_by": "t.deployed_by as deployed_by",
        "pod_id": "t.pod_id as pod_id",
    }
    sql_ORM.update(Base_object_ORM.metadata)

    def __init__(self, sc: Security_context, id=-1, metadata_id=None, user_id=-1):
        self.default_value = {
            "public": False,
            "workflow_state": "UNDEPLOYED",
            "deployed_at": None,
            "deployed_by": None,
            "pod_id": None,
        }
        self.id = id
        self.sctx = sc
        self.create_mapping(self.sql_ORM, "deployment")
        if metadata_id is not None:
            self._load_from_metadata_id(sc, metadata_id, user_id=user_id)
            # logger.info("Created dashboard with metadata_id: " + str(metadata_id))
        elif id != -1:
            self._load_from_id(sc, self.id)
            # logger.info("Created dashboard object with id: " + str(self.id))



class ImageMask:
  def __init__(self,width=-1,height=-1, polygons=[], url=None, style=None):
    self.kind = 'image-mask'
    self.width = width
    self.height = height,
    self.polygons = polygons,
    self.url = url
    self.style = style

  def to_dict(self):
    return {
      'kind': self.kind,
      'width': self.width,
      'height': self.height,
      'polygons': self.polygons,
      'url': self.url,
      'style': self.style # a JSON string containing line_color, point_color and surface_color.
    }



import logging
import os

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
LOCAL_LOG = ROOT_DIR + "/mirmod.log"


class Logger:
    def __init__(self):
        self.logger = self.get_local_logger()

    @staticmethod
    def get_local_logger():
        # Create custom logger with the specified log level
        logger = logging.getLogger("mirmod")
        log_level = logging._nameToLevel[
            os.getenv("MIRMOD_LOG_LEVEL", default="NOTSET")
        ]
        logger.setLevel(log_level)

        # Define format for logs
        log_format = "%(asctime)s - [%(levelname)s]: %(message)s"

        # Create stdout handler for logging to the console
        stdout_handler = logging.StreamHandler()
        stdout_handler.setFormatter(logging.Formatter(log_format))

        # Create file handler for logging to a file
        log_filename = os.getenv("MIRANDA_LOGFILE", default=LOCAL_LOG)
        file_handler = logging.FileHandler(log_filename)
        file_handler.setFormatter(logging.Formatter(log_format))

        # Add both handlers to the logger
        logger.addHandler(stdout_handler)
        logger.addHandler(file_handler)

        return logger

    def info(self, msg, *args, **kwargs):
        self.logger.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.logger.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.logger.error(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.logger.debug(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.logger.critical(msg, *args, **kwargs)


# We initialize a single instance of the logger here that is used by all modules
logger = Logger()



from ..security.security_context import Security_context
from ..orm.base_orm import Base_object_ORM


class Compute_policy(Base_object_ORM):
    """
    This is a WOB that defines the compute policy for a knowledge object.
    cr_group_id: The compute resource group id from which the compute scheduler will select a compute resource
    docker_image_id: The docker image id to be used for the compute resource
    """

    sql_compute_policy_ORM = {
        "id": "t.id as id",
        "order": "t.`order` as `order`",
        "cr_group_id": "t.`cr_group_id` as `cr_group_id`",
        "docker_image_id": "t.`docker_image_id` as `docker_image_id`",
        "requested_gpus": "t.`requested_gpus` as `requested_gpus`",
        "requested_cpus": "t.`requested_cpus` as `requested_cpus`",
        "requested_memory": "t.`requested_memory` as `requested_memory`",
        "bind_http": "t.`bind_http` as `bind_http`",
    }
    sql_compute_policy_ORM.update(Base_object_ORM.metadata)

    def __init__(self, sc: Security_context, id=-1, metadata_id=None, user_id=-1):
        self.default_value = {
            "order": 0,
            "cr_group_id": -1,
            "docker_image_id": -1,
            "bind_http": 1,
            "requested_gpus": 0,
            "requested_cpus": 1.0,
            "requested_memory": 2.0,
        }
        self.id = id
        self.sctx = sc
        self.create_mapping(self.sql_compute_policy_ORM, "compute_policy")
        if metadata_id is not None:
            self._load_from_metadata_id(sc, metadata_id, user_id=user_id)
        elif id != -1:
            self._load_from_id(sc, self.id)



from ..orm.base_orm import Base_object_ORM
from ..security.security_context import Security_context


class Dashboard(Base_object_ORM):
    """The dashboard is a collection of UIViews and other reports that can provide a quick
    overview of the project. The main attribute is a layout field which provide information to the client
    on how to represent the UIViews."""

    sql_ORM = {
        "id": "t.id as id",
        "layout": "t.layout as layout",
        "workflow_state": "t.workflow_state as workflow_state",
        "type": "t.type as `type`",
    }
    sql_ORM.update(Base_object_ORM.metadata)

    def __init__(self, sc: Security_context, id=-1, metadata_id=None, user_id=-1):
        self.default_value = {"order": 0, "layout": "{}", "type": "DASHBOARD"}
        self.id = id
        self.sctx = sc
        self.create_mapping(self.sql_ORM, "dashboard")
        if metadata_id is not None:
            self._load_from_metadata_id(sc, metadata_id, user_id=user_id)
            # logger.info("Created dashboard with metadata_id: " + str(metadata_id))
        elif id != -1:
            self._load_from_id(sc, self.id)
            # logger.info("Created dashboard object with id: " + str(self.id))



import json
import os
import time
import mysql.connector
import mysql.connector.pooling
from pathlib import Path
from hashlib import sha256
from mirmod.utils import logger
import uuid


def get_config(system_path="", config_file_name="config.json", ignore_env=False):
    # TODO make singleton
    if not ignore_env and config_file_name == "config.json":
        if "MIRANDA_CONFIG_JSON" in os.environ:
            try:
                config = json.loads(os.environ["MIRANDA_CONFIG_JSON"])
                logger.debug(
                    "Loaded config file from environment variable MIRANDA_CONFIG_JSON"
                )
                return config
            except Exception as e:
                logger.error(
                    "Can't load config file from environment variable MIRANDA_CONFIG_JSON"
                )
                logger.error(e)
                pass

    path = ""
    paths = [Path("/etc/miranda"), system_path, Path.home(), Path("/miranda")]
    for p in paths:
        try:
            path = p
            config_file = open(os.path.join(p, config_file_name))
            config = json.load(config_file)
            config_file.close()
            return config
        except FileNotFoundError:
            logger.debug(
                f'Tried to find config file "{config_file_name}" in location "{path}" and failed.'
            )
            pass
    logger.error(
        f"Can't find the Miranda configuration file {config_file_name} which should reside in one of the following directories: {paths}"
    )
    logger.error(f"For more information see get_config() in {__file__}")
    exit(-1)


def _create_db_cred(user, auth):
    db_username = "miranda_" + user
    m = sha256()
    db_password = m.update(b"miranda" + bytes(auth, "utf-8"))
    db_password = m.hexdigest()
    if db_username and db_password:
        return (db_username, db_password)
    else:
        logger.error("error creating database credentials")
        return None


class AuthenticationError(Exception):
    def __init__(self, n, p):
        self.name = n
        self.password = p

    def __str__(self):
        return "Authentication error: Can't authenticate as '{}' using password '{}'".format(
            self.name, self.password
        )


class Security_context:
    """The security context handles the database connection and application level identity of the API operator"""

    def __init__(
        self,
        user_name="",
        auth_string="",
        auth_from_config=False,
        system_path="",
        config_file="",
        temp_token=None,
        auth_string_is_already_hashed=False,
        pool_size: int = None,
    ):
        self.connection = None
        self.pool = None
        self.pool_size = pool_size
        self.auth_from_config = auth_from_config
        self.temp_token = temp_token
        self.last_renew = 0
        self.db_config = get_config(system_path)
        # The require_admin attribute is used to determine if a
        # a restricted SQL view is required to access data
        # or if SQL admin privileges can be assumed.
        # Note: setting require_admin = True doesn't grant admin
        # privileges
        self.require_admin = False
        if auth_from_config:
            self.application_user = self.db_config["user"]
            self.application_authstr = self.db_config["password"]
            self.database_user, self.database_password = (
                self.application_user,
                self.application_authstr,
            )
            self.require_admin = (
                True  # If we read the admin password this is an admin user
            )
        else:
            self.application_user = user_name
            self.application_authstr = auth_string
            if temp_token is None:
                self.database_user, self.database_password = _create_db_cred(
                    self.application_user, self.application_authstr
                )
                if auth_string_is_already_hashed:
                    self.database_password = self.application_authstr
            else:
                if " " in temp_token:
                    self.application_user = temp_token.split(" ")[0]
                    temp_token = temp_token.split(" ")[1]

                if temp_token.startswith("pxy."):
                    [_, user, pwd] = temp_token.split(".")
                    self.application_user = "pxy." + user
                    self.database_user, self.database_password = ("pxy." + user, pwd)
                else:
                    self.database_user, _ = _create_db_cred(
                        self.application_user, self.application_authstr
                    )
                    self.database_password = temp_token
            self.db_config["user"] = self.database_user
            self.db_config["password"] = self.database_password
        self.db_config["use_pure"] = True
        self.id = -1
        logger.debug(f"Security context created: {id(self)}")

    def __del__(self):
        logger.debug(f"Security context closed: {id(self)}")
        try:
            self.close()
        except Exception as e:
            logger.error(f"Error closing security context: {e}")

    def renew_id(self):
        if time.time() - self.last_renew < 30:
            return self.id

        self.last_renew = time.time()

        with self.connection.cursor() as cursor:
            sql = "SELECT id FROM v_user"
            cursor.execute(sql)
            res = cursor.fetchone()
            if res is None:
                self.id = -1
                # special case for when you try to authenticate using a non-miranda mysql user
                if self.auth_from_config:
                    return -1
                raise AuthenticationError(
                    self.application_user, self.application_authstr
                )
            self.id = res[0]

            if self.application_user.startswith("pxy."):
                sql = "CALL sp_extend_proxy_account_claim(%s)"
                cursor.execute(
                    sql,
                    (
                        os.environ["MIRANDA_APPLICATION_NAME"]
                        if "MIRANDA_APPLICATION_NAME" in os.environ
                        else "miranda",
                    ),
                )

        return self.id

    def user_id(self):
        if self.id == -1:
            self.id = self.renew_id()
        return self.id

    def connect(self):
        if self.pool_size is not None:
            if self.pool is None:
                pool_config = {
                    "pool_name": str(uuid.uuid4()),
                    "pool_size": self.pool_size,
                    "pool_reset_session": True,
                }
                self.pool = mysql.connector.pooling.MySQLConnectionPool(
                    **pool_config, **self.db_config
                )
            try:
                self.connection = self.pool.get_connection()
            except Exception as e:
                logger.error(f"Failed to get connection from pool: {e}")
                retry_count = 3
                while retry_count > 0:
                    try:
                        self.pool.add_connection()
                        self.connection = self.pool.get_connection()
                        logger.debug(
                            f"Added connection to pool and got connection: {self.connection}"
                        )
                        break
                    except Exception as retry_error:
                        retry_count -= 1
                        logger.error(f"Retry {3 - retry_count}/3 failed: {retry_error}")
                        if retry_count == 0:
                            raise Exception(
                                f"Failed to add and get connection after 3 retries: {retry_error}"
                            ) from e
            self.renew_id()
            return self.connection
        else:
            if self.connection is None:
                self.connection = mysql.connector.connect(**self.db_config)
                self.renew_id()
            retry = 3
            while not self.connection.is_connected() and retry > 0:
                self.connection = mysql.connector.connect(**self.db_config)
                self.renew_id()
                retry -= 1
            return self.connection

    def whoami(self):
        return (self.application_user, self.database_user)

    def close(self):
        if self.connection is not None:
            if self.connection.is_connected():
                self.connection.close()
        if self.pool is not None:
            self.pool.closeall()



class Notice:
  def __init__(self, level='info', message=''):
    self.kind = 'notice'
    self.level = level
    self.message = message

  def to_dict(self):
    return {
      'kind': self.kind,
      'level': self.level,
      'message': self.message
    }



from ..security.security_context import Security_context
from ..orm.base_orm import Base_object_ORM


class Docker_job(Base_object_ORM):
    """A docker job is created and linked to a wob which is currently executing code in a docker container.
    workflow states:
        UNINITIALIZED - The job doesn't exist.
        RUNNING       - The job is currently running in a docker container on a compute host.
        EXITED        - The job is finished and the result is available in the docker container.
        ERROR         - Something went wrong and job might not exist anymore.
    """

    docker_job_orm = {
        "id": "t.id as id",
        "host": "t.host as host",
        "container_id": "t.container_id as container_id",
        "ssh_user": "t.ssh_user as ssh_user",
        "docker_network": "t.docker_network as docker_network",
        "docker_env_vars": "t.docker_env_vars as docker_env_vars",
        "port": "t.port as port",
        "user_id": "t.user_id as user_id",
        "crg_id": "t.crg_id as crg_id",
        "workflow_state": "t.workflow_state as workflow_state",
        "notes": "t.notes as notes",
        "logs": "t.logs as logs",
        "message_id": "t.message_id as message_id",
        "tag": "t.tag as tag",
        "gpu_capacity": "t.gpu_capacity as gpu_capacity",
        "cpu_seconds": "t.cpu_seconds as cpu_seconds",
        "current_cpu": "t.current_cpu as current_cpu",
        "cpu_capacity": "t.cpu_capacity as cpu_capacity",
        "ram_gb_seconds": "t.ram_gb_seconds as ram_gb_seconds",
        "current_ram_gb": "t.current_ram_gb as current_ram_gb",
        "ram_gb_capacity": "t.ram_gb_capacity as ram_gb_capacity",
        "net_rx_gb": "t.net_rx_gb as net_rx_gb",
        "current_net_rx_gb": "t.current_net_rx_gb as current_net_rx_gb",
        "net_tx_gb": "t.net_tx_gb as net_tx_gb",
        "current_net_tx_gb": "t.current_net_tx_gb as current_net_tx_gb",
        "total_cost": "t.total_cost as total_cost",
        "is_deployed": "t.is_deployed as is_deployed",
    }
    docker_job_orm.update(Base_object_ORM.metadata)

    def __init__(self, sc: Security_context, id=-1, metadata_id=None, user_id=-1):
        self.default_value = {
            "order": 0,
            "host": "",
            "container_id": "",
            "ssh_user": "",
            "docker_network": "",
            "docker_env_vars": "",
            "workflow_state": "UNINITIALIZED",
            "message_id": -1,
            "port": 22,
            "tag": -1,
            "user_id": -1,
            "crg_id": 1,
            "gpu_capacity": 0.0,
            "cpu_seconds": 0.0,
            "current_cpu": 0.0,
            "cpu_capacity": 0.0,
            "ram_gb_seconds": 0.0,
            "current_ram_gb": 0.0,
            "ram_gb_capacity": 0.0,
            "net_rx_gb": 0.0,
            "current_net_rx_gb": 0.0,
            "net_tx_gb": 0.0,
            "current_net_tx_gb": 0.0,
            "total_cost": 0.0,
            "is_deployed": 0,
        }
        self.id = id
        self.sctx = sc
        self.create_mapping(self.docker_job_orm, "docker_job")
        if metadata_id is not None:
            self._load_from_metadata_id(sc, metadata_id, user_id=user_id)
            # logger.info("Loaded docker_job with metadata_id: " + str(metadata_id))
        elif id != -1:
            self._load_from_id(sc, self.id)
            # logger.info("Loaded docker_job with id: " + str(self.id))



from io import BytesIO, SEEK_SET, SEEK_END
import re
import pathlib
from typing import Optional
import urllib
import requests
import mimetypes

from mirmod import miranda
from mirmod.storage import StorageInterface


class ResponseStream(object):
    def __init__(self, request_iterator):
        self._bytes = BytesIO()
        self._iterator = request_iterator

    def _load_all(self):
        self._bytes.seek(0, SEEK_END)
        for chunk in self._iterator:
            self._bytes.write(chunk)

    def _load_until(self, goal_position):
        current_position = self._bytes.seek(0, SEEK_END)
        while current_position < goal_position:
            try:
                current_position += self._bytes.write(next(self._iterator))
            except StopIteration:
                break

    def tell(self):
        return self._bytes.tell()

    def read(self, size=None):
        left_off_at = self._bytes.tell()
        if size is None:
            self._load_all()
        else:
            goal_position = left_off_at + size
            self._load_until(goal_position)

        self._bytes.seek(left_off_at)
        return self._bytes.read(size)

    def seek(self, position, whence=SEEK_SET):
        if whence == SEEK_END:
            self._load_all()
        else:
            self._bytes.seek(position, whence)


class VaultStorageInterface(StorageInterface):
    def __init__(
        self,
        sctx: miranda.Security_context,
        storage_policy: miranda.Storage_policy,
        api_base_url: Optional[str] = None,
        verify_ssl: bool = True,
    ):
        super().__init__(sctx, storage_policy)
        if api_base_url is not None:
            self.api_base_url = api_base_url
        else:
            if "api_base" in self.details and self.details["api_base"] != "":
                self.api_base_url = self.details["api_base"]
            else:
                self.api_base_url = "https://miranda.localhost/api/vault/v1"

        self.verify_ssl = verify_ssl
        if (
            "access_key_id" not in self.details
            or "secret_access_key" not in self.details
        ):
            raise ValueError(
                "Vault storage policy must have access_key_id and secret_access_key set"
            )

    def _api_url(self, path: str, query: Optional[dict[str, str]] = None) -> str:
        if query is None:
            return f"{self.api_base_url}{path}"
        else:
            return f"{self.api_base_url}{path}?{urllib.parse.urlencode(query)}"

    def _api_token(self) -> str:
        return f"{self.sctx.database_user}.{self.sctx.database_password}"

    def get_vaults(self) -> list[dict[str, str]]:
        resp = requests.get(
            self._api_url("/"), cookies={"miranda_auth": self._api_token()}
        )
        resp.raise_for_status()
        return resp.json()

    def get_file_url(self, identifier: str | int) -> str:
        # TODO: add "public" flag to get public URL without secret_access_key
        return f"{self.api_base_url}/{self.details['access_key_id']}/objects/{identifier}?secret_access_key={self.details['secret_access_key']}"

    def metadata(self, identifier: str | int) -> dict[str, str]:
        resp = requests.get(
            self._api_url(
                f"/{self.details['access_key_id']}/objects/{identifier}/metadata",
                {"secret_access_key": self.details["secret_access_key"]},
            ),
            cookies={"miranda_auth": self._api_token()},
        )
        resp.raise_for_status()
        return resp.json()

    # default stream in 512kb chunks
    def download(
        self,
        identifier: str | int,
        output_filename: Optional[pathlib.Path] = None,
        to_buffer: bool = False,
        overwrite: bool = False,
        stream_chunk_size: int = 1024 * 512,
    ) -> pathlib.Path | ResponseStream:
        # if output is None, return a BytesIO
        if isinstance(identifier, str):
            object_url = self._api_url(
                f"/{self.details['access_key_id']}/objects/0",
                {
                    "secret_access_key": self.details["secret_access_key"],
                    "tag": identifier,
                },
            )
        else:
            object_url = self._api_url(
                f"/{self.details['access_key_id']}/objects/{identifier}",
                {"secret_access_key": self.details["secret_access_key"]},
            )
        cookies = {"miranda_auth": self._api_token()}
        if to_buffer:
            resp = requests.get(
                object_url, cookies=cookies, stream=True, verify=self.verify_ssl
            )
            return ResponseStream(resp.iter_content(chunk_size=stream_chunk_size))
        else:
            resp = requests.get(
                object_url, cookies=cookies, stream=True, verify=self.verify_ssl
            )
            resp.raise_for_status()

            if output_filename is not None:
                output = output_filename
            elif "Content-Disposition" in resp.headers:
                output = re.findall(
                    r'filename="(.+)"', resp.headers["Content-Disposition"]
                )[0]
            else:
                output = identifier

            output = pathlib.Path(self.storage_policy.mount_point).joinpath(output)
            output.parent.mkdir(
                parents=True, exist_ok=True
            )  # create parent directories if they don't exist

            if output.exists() and overwrite is False:
                raise FileExistsError(
                    f"File {output} already exists. Pass overwrite=True to overwrite it."
                )
            with output.open("wb") as f:
                for chunk in resp.iter_content(chunk_size=stream_chunk_size):
                    f.write(chunk)
            return output

    def upload(
        self,
        file: pathlib.Path | BytesIO,
        filename: Optional[str] = None,
        tag: Optional[str] = None,
    ) -> int | str:
        if filename is None:
            filename = file.name
        query = {"secret_access_key": self.details["secret_access_key"]}
        if tag is not None:
            query["tag"] = tag
        object_url = self._api_url(f"/{self.details['access_key_id']}/objects", query)
        cookies = {"miranda_auth": self._api_token()}

        if isinstance(file, BytesIO):
            file.seek(0)
            if filename is None:
                filename = "file"
            stream = file
            mime = "application/octet-stream"
        else:
            stream = file.open("rb")
            mime = mimetypes.guess_type(file)[0]
        # upload with body byte stream
        resp = requests.post(
            object_url,
            cookies=cookies,
            data=stream,
            headers={"Content-Type": mime, "x-filename": filename},
            verify=self.verify_ssl,
        )
        resp.raise_for_status()
        return resp.json()["oid"]



import datetime
import json
from lxml import etree
import networkx as nx
from mirmod import _table_to_object_lookup, miranda
import traceback


def import_graph(ko, graphml_file):
    G = nx.parse_graphml(graphml_file)
    nodes = {}
    wobs = {}
    for node in G.nodes(data=True):
        if "class" in node[1]:
            if node[1]["class"].upper() not in _table_to_object_lookup:
                raise Exception(f"Unknown class {node[1]['class'].upper()}")
            if node[1]["class"].upper() != "KNOWLEDGE_OBJECT":
                nodes[node[0]] = node[1]
        else:
            raise Exception(f'No node class type found for node "{node[0]}"')

    for node in nodes:
        print(f"Creating {node}")
        node_data = nodes[node]
        node_class = node_data["class"].upper()
        print(f"  class: {node_class}")
        try:
            if node_class == "CODE" or node_class == "CODE_BLOCK":
                wob = miranda.create_wob(
                    ko,
                    name=node_data["name"] if "name" in node_data else "Unnamed Code",
                    description=node_data["description"]
                    if "description" in node_data
                    else "",
                )
                if wob.id == -1:
                    raise Exception("Failed to create code block")
                for attr in node_data:
                    if attr in ["name", "description", "id", "metadata_id"]:
                        continue
                    print("  Setting attribute {} to {}".format(attr, node_data[attr]))
                    if attr in wob.orm.keys():
                        try:
                            if attr in wob.default_value.keys():
                                if isinstance(wob.default_value[attr], str):
                                    setattr(wob, attr, node_data[attr])
                                elif isinstance(wob.default_value[attr], int):
                                    setattr(wob, attr, int(node_data[attr]))
                                elif isinstance(wob.default_value[attr], float):
                                    setattr(wob, attr, float(node_data[attr]))
                            else:
                                setattr(wob, attr, node_data[attr])
                        except Exception as e:
                            print(f"  Error setting attribute {attr} : {e}")
                    else:
                        wob.set_attribute(attr, node_data[attr])
                wobs[node] = wob
                wob.update(ko.sctx)
        except Exception as e:
            print(f"  Error: {e}")
            print(traceback.format_exc())

    for edge in G.edges(data=True):
        print(f"Creating edge {edge[0]} -> {edge[1]}, data= {edge[2]}")
        try:
            src = wobs[edge[0]]
            dst = wobs[edge[1]]
            print(
                f"Translated to wob mids: edge {src.metadata_id} -> {dst.metadata_id}"
            )
            attr: dict = json.loads(edge[2]["attributes"])
            for dest_socket in attr.keys():
                src_socket = attr[dest_socket]
                print(f"  Linking {src_socket} to {dest_socket}")
                try:
                    if miranda.link(
                        ko.sctx,
                        src,
                        dst,
                        src_socket[0],
                        dest_socket,
                        datatype=src_socket[1],
                    ):
                        print("Error while linking")
                except Exception as e:
                    print(f"  Error: {e}")
                    print(traceback.format_exc())

        except Exception as e:
            print(f"  Error: {e}")
            print(traceback.format_exc())


def export_graph(graph):
    root = etree.Element("graphml")
    root.set("xmlns", "http://graphml.graphdrawing.org/xmlns")
    root.set("__xmlnsxsi__", "http://www.w3.org/2001/XMLSchema-instance")
    root.set("__xmlnsy__", "http://www.yworks.com/xml/graphml")
    root.set(
        "__xsischemaLocation__",
        "http://graphml.graphdrawing.org/xmlns http://www.yworks.com/xml/schema/graphml/1.1/ygraphml.xsd",
    )
    root.append(etree.Comment(f"Generated by Miranda on {datetime.datetime.now()}"))
    gr = etree.SubElement(root, "graph")
    gr.set("edgedefault", "directed")
    key_map = []
    node_els = []
    for id, node in graph.nodes(data=True):
        if "wob" not in node:
            continue
        node_el = etree.SubElement(gr, "node")
        node_el.set("id", f"n{id}")
        node_el.set("attr.name", node["wob"]["name"])
        for key in node["wob"]:
            key_i = key_map.index(key) if key in key_map else None
            if key_i is None:
                key_i = len(key_map)
                key_map.append(key)
                key_el = etree.SubElement(root, "key")
                key_el.set("id", f"d{key_i}")
                key_el.set("for", "node")
                key_el.set("attr.name", key)
                if isinstance(node["wob"][key], int):
                    key_el.set("attr.type", "int")
                elif isinstance(node["wob"][key], float):
                    key_el.set("attr.type", "double")
                else:
                    key_el.set("attr.type", "string")
            data_el = etree.SubElement(node_el, "data")
            data_el.set("key", f"d{key_i}")
            data_el.text = str(node["wob"][key]).replace("\r\n", "\n")
        node_els.append(node_el)
        gr.append(node_el)

    key_map.append("nodegraphics")
    ngkeyi = len(key_map) - 1
    ngkey_el = etree.SubElement(root, "key")
    ngkey_el.set("id", f"d{ngkeyi}")
    ngkey_el.set("for", "node")
    ngkey_el.set("yfiles.type", "nodegraphics")

    for node in node_els:
        data_el = etree.SubElement(node, "data")
        data_el.set("key", f"d{ngkeyi}")
        shape_el = etree.SubElement(data_el, "__yShapeNode__")
        label_el = etree.SubElement(shape_el, "__yNodeLabel__")
        label_el.text = node.get("attr.name")
        geometry_el = etree.SubElement(shape_el, "__yGeometry__")
        geometry_el.set("height", "30.0")
        geometry_el.set("width", f"{len(label_el.text) * 6.0 + 10.0}")

    id = 0
    for src, dest, edge in graph.edges(data=True):
        edge_el = etree.SubElement(gr, "edge")
        edge_el.set("source", f"n{src}")
        edge_el.set("target", f"n{dest}")
        data = etree.SubElement(edge_el, "data")
        data.set("key", "e{}".format(id))
        data.text = json.dumps(edge["attributes"])
        key = etree.SubElement(root, "key")
        key.set("id", f"e{id}")
        key.set("for", "edge")
        key.set("attr.name", "attributes")
        key.set("attr.type", "string")
        id += 1
        gr.append(edge_el)
        root.append(key)

    treestr = etree.tostring(root, pretty_print=True)
    # stupid hack to get past the lxml limitation on not using : in tag and attribute names
    treestr = treestr.replace(b"__yShapeNode__", b"y:ShapeNode")
    treestr = treestr.replace(b"__yNodeLabel__", b"y:NodeLabel")
    treestr = treestr.replace(b"__yGeometry__", b"y:Geometry")
    treestr = treestr.replace(b"__xmlnsxsi__", b"xmlns:xsi")
    treestr = treestr.replace(b"__xmlnsy__", b"xmlns:y")
    treestr = treestr.replace(b"__xsischemaLocation__", b"xsi:schemaLocation")
    return treestr



from typing import Callable, Dict
import functools


class FailureMode(Exception):
    def __init__(self, mode: str, original_exception: Exception):
        self.mode = mode
        self.original_exception = original_exception
        super().__init__(
            f"|=> ERROR: Failure mode '{mode}' occurred: {str(original_exception)}"
        )


class RecoveryStrategy:
    def __init__(self, strategy: Callable):
        self.strategy = strategy

    def execute(self, *args, **kwargs):
        return self.strategy(*args, **kwargs)


class FailureModeHandler:
    def __init__(self, failure_modes: Dict[str, RecoveryStrategy]):
        self.failure_modes = failure_modes

    def __call__(self, func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                for mode, recovery_strategy in self.failure_modes.items():
                    if mode.lower() in str(e).lower():
                        if mode.startswith("|=>"):
                            mode = mode[4:]
                        print(f"|=> Attempting to recover from failure mode: {mode}")
                        try:
                            result = recovery_strategy.execute(*args, **kwargs)
                            print(f"|=> Recovery successful for {mode}")
                            return result
                        except Exception as recovery_error:
                            err = str(recovery_error)
                            err = err.replace("|=>", "")
                            print(f"|=> Recovery failed for {mode}: {err}")
                            raise FailureMode(mode, e)
                raise FailureMode("|=> ERROR: Unexpected failure", e)

        return wrapper
